/*
 * Copyright (C) 2013-2014 SEH Computertechnik GmbH
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <linux/kthread.h>
#include <linux/uaccess.h>
#include <linux/slab.h>
#include <linux/utsname.h>


#include "vhci_common.h"
#include "vhci.h"
#include "vhci_event.h"
#include "vhci_char_dev.h"
#include "vhci_to_service.h"
#include "vhci_sd_boost.h"

#define KERNEL
#define __OS_LINUX__

/*DUMMY */
#define usb_device_descriptor void*
#define vusbResponseHeader void *

#include "service_defs.h"


#define USBDEVFS_URB_TYPE_ISO		   0
#define USBDEVFS_URB_TYPE_INTERRUPT	   1
#define USBDEVFS_URB_TYPE_CONTROL	   2
#define USBDEVFS_URB_TYPE_BULK		   3

static int unused __attribute__((unused));
void vhci_device_unlink_all(struct vhci_device *vdev);

static inline void seh_vusb_get_dev_ids (__u32 *bus, __u32 *addr, __u32 *ip, vusb_device_addr *indexes){
    *bus = indexes->busno;
    *addr = indexes->addr;
    *ip = indexes->ipl;
}


/*
 *  seh_vusb_plug_hw
 *
 *
 *
 *
*/
int32_t seh_vusb_plug_hw (vusb_plugin_hardware *hw){
	uint32_t rhport = 0, speed = 0, addr, ip, bus, n, hcd_index, hcd_end;
	struct vhci_device *vdev, *vdev_free = NULL;
    struct vhci_hcd *hcd_controller, *hcd_controller_free = NULL;
    
    seh_dbg_vhci_service("Enter\n");
	seh_vusb_get_dev_ids (&bus, &addr, &ip, &hw->addr); 
    
    if (hw->bcdUSB >= 0x0300){
        speed  = USB_SPEED_SUPER;
        hcd_index = VHCI_FIRST_USB3;
        hcd_end = VHCI_CONTROLLERS;
    }else{
	if (hw->bcdUSB >= 0x0200)
		speed  = USB_SPEED_HIGH;
	else if (hw->bcdUSB >= 0x0110)
		speed  = USB_SPEED_FULL;
	else if (hw->bcdUSB >= 0x0100)
		speed  = USB_SPEED_LOW;
        hcd_index = 0;
        hcd_end = VHCI_FIRST_USB3;
    }    
    for (; hcd_index < hcd_end; hcd_index++)
    {
        hcd_controller = controllers[hcd_index];
        /*check if not assigned by any other and store first free*/
        for (rhport = 0; rhport < VHCI_NPORTS; rhport++){
            vdev = port_to_vdev(hcd_controller, rhport);
            //seh_dbg_vhci_service("vdev %p\n",vdev);
            if (vdev->device_status == VDEV_ST_NULL){ 
                if (!vdev_free) 
                    vdev_free = vdev;
                if (!hcd_controller_free) 
                    hcd_controller_free = hcd_controller;
            }
            if (vdev->bus == bus && vdev->addr == addr && vdev->ip == ip && 
                vdev->device_status != VDEV_ST_NULL) {
                /* end of the lock */
                dev_err(NULL, "port %d already used(dev %d/%d/%x stat %d\n",rhport, bus,addr, ip, vdev->device_status);
                return -EINVAL;
            }
            
        }
    }
    if (!vdev_free){
        pr_err ("Root hub no free ports found\n");
        return 0;
    }
    rhport = hcd_port_index(vdev_free->rh_port_id);
    seh_dbg_vhci_service("vdev_free %p\n",vdev_free);
    seh_dbg_vhci_service("hcd_controller_free %p\n",hcd_controller_free);
    
	vdev_free->bus               = bus;
	vdev_free->addr              = addr;
	vdev_free->ip                = ip;
	vdev_free->speed             = speed;
	vdev_free->device_status     = VDEV_ST_NOTASSIGNED;
	vdev_free->useSSL            = hw->useSSL;
    vdev_free->buffersHackBoosted = 0;
    
	n = sizeof(vdev_free->username);
	strncpy(vdev_free->username, hw->username, n);
	if (n > 0)
		vdev_free->username[n-1] = '\0';

	rh_port_connect(hcd_controller_free, rhport, speed);
	
	seh_dbg_vhci_service("rhport(%u) speed(%u)\n", rhport, speed);
	/*now send to service command - a little bit stupit protocol but...*/
	
	seh_vusb_add_cmd_to_svc(hcd_controller_free->id, vdev_free, CLAIM);
	seh_vusb_add_cmd_to_svc(hcd_controller_free->id, vdev_free, IF_STAT);
    
	seh_dbg_vhci_service("Exit\n");
	return 0;
  
}
/*
 * Unplug
 */
int32_t seh_vusb_unplug_hw (vusb_plugout_hardware *hw){
	uint32_t hcd_index, rhport;
    struct vhci_device *vdev;
	bool found = false;
    
    seh_dbg_vhci_service("Enter\n");
	
	for (hcd_index = 0; hcd_index < VHCI_CONTROLLERS; hcd_index++){
        for (rhport = 0; rhport < VHCI_NPORTS; rhport++){
            vdev = port_to_vdev(controllers[hcd_index], rhport);
            if (vdev->bus == hw->addr.busno && vdev->addr == hw->addr.addr && vdev->ip == hw->addr.ipl){
                found = true;
                break; 
            }
        }
        if (found) break;
    }
    if (!found){
		pr_err ("Root hub port not found\n");
		return -EINVAL;
	}
    
	if (vdev->device_status == VDEV_ST_NULL) {
		pr_err("not connected %d\n", vdev->device_status);
		return -EINVAL;
	}
    unpluging_device(vdev);
    
	seh_dbg_vhci_service("Exit\n");
	return 0;
}
/*
 * Eject
 */
int32_t seh_vusb_eject_hw (vusb_plugout_hardware *hw){
  return seh_vusb_unplug_hw (hw);
}
/*
 * Get instance ID
 */
int32_t seh_vusb_get_inst_id (vusb_get_instance_id *hw, uint32_t *Id){
	
    uint32_t rhport = -1;
    uint32_t hcd_index;
    struct vhci_device *vdev;
    
    seh_dbg_vhci_service("Enter %s\n", __FUNCTION__);
    
    for (hcd_index = 0; hcd_index < VHCI_CONTROLLERS; hcd_index++){
        for (rhport = 0; rhport < VHCI_NPORTS; rhport++){
            vdev = port_to_vdev(controllers[hcd_index], rhport);
            if (vdev->bus == hw->addr.busno && vdev->addr == hw->addr.addr && vdev->ip == hw->addr.ipl){
                rhport = rhport_get(hcd_index, rhport);
                break;
            }
        }
        if (rhport != -1) break;
    }
    if (rhport == -1){
		pr_err ("Root hub port not found\n");
		return -EINVAL;
	}
    *Id = rhport;
    seh_dbg_vhci_service("Exit instance ID %d\n", rhport);
    return 0;
}
/*
 * Get if status
 */
int32_t seh_vusb_get_if_status(vusb_get_if_status *hw, uint32_t *Status){
	uint32_t hcd_index, rhport;
    struct vhci_device *vdev;
    bool found = false;
    seh_dbg_vhci_service("Enter\n");

    for (hcd_index = 0; hcd_index < VHCI_CONTROLLERS; hcd_index++){
        for (rhport = 0; rhport < VHCI_NPORTS; rhport++){
            vdev = port_to_vdev(controllers[hcd_index], rhport);
            if (vdev->bus == hw->addr.busno && vdev->addr == hw->addr.addr && vdev->ip == hw->addr.ipl){
                found = true;
                break; 
            }
        }
        if (found) break;
    }
    if (!found){
		*Status = PNP_NotStarted;
		seh_dbg_vhci_service("Exit status %d port rhport unknown\n", *Status);
		return 0;
	}
    /* [fl] assign corresponding windows pnp states used by utnservice */
    if (vdev->device_status == VDEV_ST_NOTASSIGNED) 
        *Status = PNP_NotStarted;
    else if (vdev->device_status == VDEV_ST_USED) 
        *Status = PNP_Started;
    /*stop pending means we wait service to read the status
     * than we make real reset*/
    else if (vdev->device_status == VDEV_ST_STOP_PENDING){
        *Status = PNP_StopPending;
    }
    else if (vdev->device_status == VDEV_ST_STOPPED){
        unplug_device(vdev);
        *Status = PNP_Deleted;
    }
    else
        *Status = PNP_Unknown;
    seh_dbg_vhci_service("Exit status %d port rhport %d\n", *Status, vdev->rh_port_id);
    return 0;
}
int32_t seh_vusb_set_hb (void){
  return 0;
}

int32_t seh_vusb_hartbeat (void) {
  seh_vusb_boost_storage_devices();
  return 0;
}

/*
 * URB processing 
 */
static inline int fill_urb_to_svc(struct vhci_priv *priv, struct urb *urb,
			      struct usb_host_endpoint	*ep, struct usb_ctrlrequest *setup,VUSB_GET_URB *pServiceUrb){
	uint32_t len = 0;
    struct _VUSB_URB_HEADER *uh = &pServiceUrb->data.header;
    void *data = NULL;
    /*Fill data*/
    __put_user (IO_TYPE_URB, &pServiceUrb->dataType );
    /*fill header*/
    __put_user (priv->vdev->addr, &uh->addr);
    __put_user (priv->vdev->bus, &uh->busnr); 
    /*__put_user (usb_pipeendpoint(urb->pipe), &uh->endpoint);*/
    __put_user (ep->desc.bEndpointAddress, &uh->endpoint);

    switch (usb_endpoint_type(&ep->desc)){
		case USB_ENDPOINT_XFER_ISOC :
			__put_user (USBDEVFS_URB_TYPE_ISO, &uh->type);
		break;
		case USB_ENDPOINT_XFER_INT:
			__put_user (USBDEVFS_URB_TYPE_INTERRUPT, &uh->type);
		break;
		case USB_ENDPOINT_XFER_CONTROL:
			__put_user (USBDEVFS_URB_TYPE_CONTROL, &uh->type);
		break;
		case USB_ENDPOINT_XFER_BULK:
			__put_user (USBDEVFS_URB_TYPE_BULK, &uh->type);
		break;
    }
    /*put max len if out*/
    if (usb_pipein(urb->pipe))
			__put_user (htonl(urb->transfer_buffer_length), &uh->maxLength);
    else
		__put_user (0, &uh->maxLength);

    len = 0;
    if (USB_ENDPOINT_XFER_CONTROL == usb_endpoint_type(&ep->desc)){
	VUSB_CTRL_TRANSFER   *rq  = &pServiceUrb->data.urbCtrlRequest.ctrl;
	__put_user (setup->bRequestType, &rq->bmRequestType );
	__put_user (setup->bRequest, &rq->bRequest );
	__put_user (htons(setup->wValue), &rq->wValue );
	__put_user (htons(setup->wIndex), &rq->wIndex );
	__put_user (htons(setup->wLength), &rq->wLength );
	__put_user ((setup->wValue>>8), &pServiceUrb->descriptorType); 
	len += sizeof (struct _VUSB_CTRL_TRANSFER);
	data =  &pServiceUrb->data.urbCtrlRequest.data;
	
	seh_dbg_vhci_data("CONTROL  \n\bReqType %d,bRequest %d, tWValue %d \n\tWIndex %d, WLength %d,DescType %d, Len %u\n", 
		    setup->bRequestType,
		 setup->bRequest, setup->wValue, setup->wIndex , setup->wLength , (setup->wValue>>8), len);
    }else{
		data = &pServiceUrb->data.urbBulkOrInterruptRequest.data;

    }
    if (!usb_pipein(urb->pipe)){
        if (!urb->num_sgs){
            if (urb->transfer_buffer_length && data){
      len += urb->transfer_buffer_length;
      unused=copy_to_user (data, urb->transfer_buffer, urb->transfer_buffer_length);
              /*seh_dbg_vhci_hc( "OUT urb->transfer_buffer %x  %x urb->transfer_buffer_length %d urb->transfer_flags %x\n", urb->transfer_buffer , urb->transfer_dma ,urb->transfer_buffer_length,
                urb->transfer_flags);*/
            }
        }else
        {
            struct sg_mapping_iter *miter = &priv->miter;
            u32 this_sg;
            bool next_sg;
            u32 trans = 0;
            u32 all_len = urb->transfer_buffer_length;
            u32 flags = SG_MITER_FROM_SG;
            sg_miter_start(miter, urb->sg, urb->num_sgs, flags);

            next_sg = sg_miter_next(miter);
            if (next_sg == false) {
                WARN_ON_ONCE(1);
                printk ("urb->num_sgs %d\n", urb->num_sgs);
                sg_miter_stop(miter);
                return -1;
            }
            do {
                this_sg = min_t(u32, all_len, miter->length);
                miter->consumed = this_sg;
                trans += this_sg;
                unused=copy_to_user (data, miter->addr, this_sg);
                all_len -= this_sg;

                if (!all_len)
                    break;
                next_sg = sg_miter_next(miter);
                if (next_sg == false) {
                    printk ("urb->num_sgs %d\n", urb->num_sgs);
                    WARN_ON_ONCE(1);
                    return -1;
                }
                data += this_sg;
            } while (1);
            sg_miter_stop(miter);
            len += urb->transfer_buffer_length;
        }
    }
    __put_user (htonl(len), &uh->length); /*only data + control pkt*/
    
    len += sizeof (struct _VUSB_URB_HEADER);
    __put_user (len, &pServiceUrb->dataLength );
    
    seh_dbg_vhci_data("Sent %s seq %u Bus %d Addr %d EP %x type %d BufferLen %d\n\t"
            "setup %p transf AllSize %u\n",
            (usb_pipein(urb->pipe) ? "IN":"OUT"),priv->seqnum,
            priv->vdev->bus, priv->vdev->addr, ep->desc.bEndpointAddress, usb_endpoint_type(&ep->desc),
            urb->transfer_buffer_length, urb->setup_packet, len
          );
    return 0;
}
static inline void fill_set_cfg_to_svc(struct vhci_priv *priv, struct urb *urb,
			      struct usb_host_endpoint	*ep, struct usb_ctrlrequest *setup,VUSB_GET_URB *pServiceUrb){
	__put_user (IO_TYPE_SET_CONFIGURATION, &pServiceUrb->dataType);
	__put_user (priv->vdev->addr, &pServiceUrb->buffer[0]);
	__put_user (priv->vdev->bus, &pServiceUrb->buffer[1]);
	__put_user ((setup->wValue), &pServiceUrb->buffer[2]);
	__put_user (0, &pServiceUrb->buffer[3]);
	__put_user (4, &pServiceUrb->dataLength);
    seh_dbg_vhci_data("Sent SET_CONFIG priv %p Num %u, urb %p\n", priv, priv->seqnum, priv->urb);
	
	seh_dbg_vhci_rx("CONTROL  \n\tReqType %d,tRequest %d, tWValue %d \n\tWIndex %d, WLength %d,DescType %d, DATA %d", 
	    setup->bRequestType,setup->bRequest, setup->wValue, setup->wIndex , 
		setup->wLength , (setup->wValue>>8), priv->urb->transfer_buffer_length);
	seh_dbg_vhci_data("Sent SET CFG PIPE Bus %d Addr %d, EP %d\n", priv->vdev->bus, priv->vdev->addr, ep->desc.bEndpointAddress);
}	
static inline void fill_reset_pipe_to_svc(struct vhci_priv *priv, struct urb *urb,
			      struct usb_host_endpoint	*ep, struct usb_ctrlrequest *setup,VUSB_GET_URB *pServiceUrb){
	__put_user (IO_TYPE_RESET_PIPE, &pServiceUrb->dataType);
	__put_user (priv->vdev->addr, &pServiceUrb->buffer[0]);
	__put_user (priv->vdev->bus, &pServiceUrb->buffer[1]);
	__put_user ((setup->wIndex), &pServiceUrb->buffer[2]);
	__put_user (3, &pServiceUrb->dataLength);
	seh_dbg_vhci_data("Sent RESET PIPE Bus %d Addr %d, EP %d\n", priv->vdev->bus, priv->vdev->addr, ep->desc.bEndpointAddress);
}
static inline void fill_abort_pipe_to_svc(struct vhci_priv *priv, struct urb *urb,
										  struct usb_host_endpoint	*ep, struct usb_ctrlrequest *setup,
										  VUSB_GET_URB *pServiceUrb){
	__put_user (IO_TYPE_ABORT_PIPE, &pServiceUrb->dataType);
	__put_user (priv->vdev->addr, &pServiceUrb->buffer[0]);
	__put_user (priv->vdev->bus, &pServiceUrb->buffer[1]);
	__put_user (ep->desc.bEndpointAddress, &pServiceUrb->buffer[2]);
	__put_user (3, &pServiceUrb->dataLength);
	 seh_dbg_vhci_data("Sent ABORT PIPE Bus %d Addr %d, EP %d\n", priv->vdev->bus, priv->vdev->addr, ep->desc.bEndpointAddress);
}

static inline void fill_unlink_to_svc(struct vhci_priv *priv, VUSB_GET_URB *pServiceUrb){
    __put_user (IO_TYPE_UNLINK_URB, &pServiceUrb->dataType);
	__put_user (priv->unlink_seqnum, (uint32_t *)&pServiceUrb->buffer[0]);
	__put_user (sizeof (uint32_t), &pServiceUrb->dataLength);
    seh_dbg_vhci_data("Sent UNLINK\n");
}
static inline void fill_claim_to_svc(struct vhci_priv *priv, VUSB_GET_URB *pServiceUrb){
	__put_user (IO_TYPE_CLAIMDEVICE, &pServiceUrb->dataType);
    __put_user ((char)priv->vdev->addr, &pServiceUrb->buffer[0]);
    __put_user ((char)priv->vdev->bus, &pServiceUrb->buffer[1]);
    __put_user (2, &pServiceUrb->dataLength);
    seh_dbg_vhci_data("Sent CLAIM priv %p addr %d bus %d\n", priv, priv->vdev->addr, priv->vdev->bus);
}
static inline void fill_claim2_to_svc(struct vhci_priv *priv, VUSB_GET_URB *pServiceUrb){
#define NAME_SIZE 32	
	int32_t name_len = 0;
	__put_user (IO_TYPE_CLAIMDEVICE2, &pServiceUrb->dataType);
    __put_user ((char)priv->vdev->addr, &pServiceUrb->buffer[0]);
    __put_user ((char)priv->vdev->bus, &pServiceUrb->buffer[1]);

	name_len = sizeof (priv->vdev->username);
	name_len = (NAME_SIZE < name_len) ? NAME_SIZE:name_len;
	unused=copy_to_user (&pServiceUrb->buffer[2], priv->vdev->username, name_len);
    __put_user (2+name_len, &pServiceUrb->dataLength);
    
	seh_dbg_vhci_data("Sent CLAIM2 priv %p addr %d bus %d\n", priv, priv->vdev->addr, priv->vdev->bus);
}
static inline void fill_ifstatus_to_svc(struct vhci_priv *priv, VUSB_GET_URB *pServiceUrb){
	__put_user (IO_TYPE_UNKNOWN, &pServiceUrb->dataType);
    __put_user ((char)priv->vdev->addr, &pServiceUrb->buffer[0]);
    __put_user ((char)priv->vdev->bus, &pServiceUrb->buffer[1]);
    __put_user (2, &pServiceUrb->dataLength);
    seh_dbg_vhci_data("Sent ISTATUS priv %p addr %d bus %d\n", priv, priv->vdev->addr, priv->vdev->bus);
}
/*
 * vhci_get_priv_send_size
 * calculate size in bytes requered for sending one URb to service
 * 
 * NOTE as there are unions - we calculate always more data than real - but ok
 */ 
int32_t vhci_get_priv_send_size(struct vhci_priv *priv){
    size_t size = 0;
    size = sizeof (VUSB_GET_URB);
    if (is_urb(priv) && priv->urb->transfer_buffer_length && !usb_pipein(priv->urb->pipe))
        size += priv->urb->transfer_buffer_length;
    else if (is_claim(priv)){
        size += NAME_SIZE + 8;
    }else
        size += 8;
      
return size;
}

/*
 * vhci_send_to_service
 * send URB to service
 */
int32_t vhci_send_to_service (struct vhci_priv *priv,size_t priv_size, uint8_t __user *buf, size_t count){
	VUSB_GET_URB *pServiceUrb = (VUSB_GET_URB *)buf;  
	struct urb *urb;
	struct usb_ctrlrequest *setup;

	urb = priv->urb;
  
	if (unlikely(!access_ok(VERIFY_WRITE, (void __user *)pServiceUrb, count))){
		pr_err("Access write fail\n");
		return 0;
	}
	__put_user (priv_size, &pServiceUrb->size);
	__put_user (0, &pServiceUrb->descriptorType); 
	__put_user (priv->vdev->bus, &pServiceUrb->busno);
	__put_user (priv->vdev->addr, &pServiceUrb->addr);
	__put_user (priv->vdev->ip, &pServiceUrb->ipl);
	__put_user (priv->seqnum, &pServiceUrb->pIoRequestHandle);
	__put_user (priv->vdev->useSSL, &pServiceUrb->useSSL);
	__put_user (0, &pServiceUrb->data.header.maxLength);

	if (is_urb(priv)){
		struct usb_host_endpoint	*ep;
		struct usb_device		*dev;

		dev = urb->dev;
		if (unlikely(!dev)){
			pr_err("no dev\n");
			return 0;
		}
		ep = (usb_pipein(urb->pipe) ? dev->ep_in : dev->ep_out)
			[usb_pipeendpoint(urb->pipe)];
		if (!ep){
			pr_err("no ep\n");
			return 0;
		}
		setup = (struct usb_ctrlrequest *) urb->setup_packet;
    
		/*check for SET configuration*/
		if (usb_endpoint_type(&ep->desc) ==  USB_ENDPOINT_XFER_CONTROL && setup){
			
			if (setup->bRequestType == 0 && setup->bRequest == USB_REQ_SET_CONFIGURATION){
				/*Set configuration request*/
				fill_set_cfg_to_svc (priv, urb, ep, setup, pServiceUrb);
			}else if (setup->bRequest == USB_REQ_CLEAR_FEATURE 
					  && setup->bRequestType == USB_RECIP_ENDPOINT 
					  && setup->wValue == USB_ENDPOINT_HALT){
				/*Reset Pipe request*/
				fill_reset_pipe_to_svc (priv, urb, ep, setup, pServiceUrb);
			} else if (setup->bRequest == USB_REQ_SET_FEATURE 
					&& setup->bRequestType == USB_RECIP_ENDPOINT 
					&& setup->wValue == USB_ENDPOINT_HALT){
				/*Abort Pipe request*/
				fill_abort_pipe_to_svc (priv, urb, ep, setup, pServiceUrb);
			}else{
				if (fill_urb_to_svc (priv, urb, ep, setup, pServiceUrb) == -1)
                    return 0;
			}
		}else{
            if (fill_urb_to_svc (priv, urb, ep, setup, pServiceUrb) == -1)
                return 0;
		}
	}else if (is_unlink(priv)){
		fill_unlink_to_svc (priv, pServiceUrb);
	}else if (is_claim(priv)){
		fill_claim2_to_svc (priv, pServiceUrb);
	}else if (is_ifstatus(priv))
		fill_ifstatus_to_svc (priv, pServiceUrb);
	return priv_size;
}


inline bool get_indexes_from_buf (const void __user *buf, size_t count, __u32 *bus, __u32 *addr, __u32 *ip, __u32 *seqnum){
	VUSB_SET_URB service_urb;
  
	if (unlikely(!access_ok(VERIFY_READ, (void __user *)buf, sizeof (VUSB_SET_URB)))){
		pr_err("Cannot read user data count %zu\n", sizeof (VUSB_SET_URB));
		return false;
	}

	if (unlikely(unused=__copy_from_user(&service_urb, (void __user *)buf, sizeof (VUSB_SET_URB)))){
		pr_err("Copy from user fail\n");
		return false;
	}
  
	if (unlikely(service_urb.size != count)){
		pr_err ("Bad incomming URB, data size %u vs %zu\n", service_urb.size, count);
		return false;
	}
	*bus = service_urb.busno;
	*addr = service_urb.addr;
	*ip = service_urb.ipl;
	*seqnum = service_urb.pIoRequestHandle;
        
	return true;
}
/*
 * Get data from Serivce
 */
int32_t vhci_get_from_service (struct vhci_priv *priv, const void __user *buf, size_t count, struct urb *urb){
	VUSB_SET_URB __user *pServiceUrb = (VUSB_SET_URB *)buf; 
	int32_t error;
	int32_t status;   
    int32_t dataLength;
	
	seh_dbg_vhci_rx("Enter buf %p , size %zu, urb %p\n", buf, count, urb);
	
	if (unlikely(!access_ok(VERIFY_READ, (void __user *)buf, count))){
		pr_err("Cannot read user data count %zu\n", count);
		return 0;
	}
	
	__get_user(error, &pServiceUrb->error);
    __get_user(status, &pServiceUrb->status);
    __get_user(dataLength, &pServiceUrb->dataLength);
    
	if (unlikely(status == StatSysErr || status == StatAbortPipe)){
        seh_dbg_vhci_rx("ERROR: error %u, status %u\n", error, status);
		if (urb->status == 71)/*no idea why*/
			urb->status = -EPROTO;
		else
			urb->status = -error;
	}else
		urb->status = status;
  
    if (unlikely(error != 0 || status != 0)){
        seh_dbg_vhci_rx("ERROR: error %u, status %u\n", error, status);
	}
    if (usb_pipein(urb->pipe) && dataLength){
		if (urb->transfer_buffer_length >= dataLength){
		    if (!urb->num_sgs){
		        if (urb->transfer_buffer && pServiceUrb->data){
		            /*seh_dbg_vhci_hc( "IN urb->transfer_buffer %p  %p urb->transfer_buffer_length %d urb->transfer_flags %x\n", urb->transfer_buffer , urb->transfer_dma ,urb->transfer_buffer_length,
		            urb->transfer_flags);*/
		    unused = __copy_from_user (urb->transfer_buffer, (void __user *)pServiceUrb->data, dataLength);
		        }
		    }else
		    {
		        struct sg_mapping_iter *miter = &priv->miter;
		        u32 this_sg;
		        bool next_sg;
		        u32 trans = 0;
		        u32 all_len = dataLength;
		        void __user *ubuf = pServiceUrb->data;
                u32 flags = SG_MITER_TO_SG;
		        sg_miter_start(miter, urb->sg, urb->num_sgs, flags);

		        next_sg = sg_miter_next(miter);
		        if (next_sg == false) {
		            WARN_ON_ONCE(1);
                    pr_err ("urb->num_sgs %d\n", urb->num_sgs);
		            return 0;
		        }
		        do {
		            this_sg = min_t(u32, all_len, miter->length);
		            miter->consumed = this_sg;
		            trans += this_sg;
		            unused = __copy_from_user (miter->addr, ubuf, this_sg);
		            all_len -= this_sg;
                    if (!all_len){
		                break;
                    }
		            next_sg = sg_miter_next(miter);
		            if (next_sg == false) {
		                WARN_ON_ONCE(1);
                        pr_err ("urb->num_sgs %d\n", urb->num_sgs);
		                return 0;
		            }

		            ubuf += this_sg;
		        } while (1);
		        sg_miter_stop(miter);
		    }
		}else{
		    pr_err("ERROR: data len %d, buffer len %u status %d\n", urb->transfer_buffer_length, dataLength, urb->status);
		}
    }
    if (usb_pipein(urb->pipe)){
        urb->actual_length = dataLength;
	}else{
		urb->actual_length = urb->transfer_buffer_length;
    }
	seh_dbg_vhci_rx("Exit Receve from Service URB Len %d, status %d\n", urb->actual_length, urb->status);
	return 0;
}

int32_t vhci_get_cmd_reply_from_service (const void __user *buf, size_t count){
	VUSB_SET_URB __user *pServiceUrb = (VUSB_SET_URB *)buf; 
	int32_t error;
	int32_t status;
	
	seh_dbg_vhci_rx("Enter buf %p , size %zu\n", buf, count);
	
	if (unlikely(!access_ok(VERIFY_READ, (void __user *)buf, count))){
		pr_err("Cannot read user data count %zu\n", count);
		return 0;
	}
	
	__get_user(error, &pServiceUrb->error);
	__get_user(status, &pServiceUrb->status);
    seh_dbg_vhci_rx("Status %x error %x\n", status, error);
	return status;
}

inline int32_t get_size_from_buf (uint8_t __user *buf, uint32_t *hcd_index, size_t *size){
    VUSB_GET_URB *pServiceUrb = (VUSB_GET_URB *)buf;
    
    if (unlikely(!access_ok(VERIFY_READ, (void __user *)&pServiceUrb->size, sizeof (pServiceUrb->size)))){
		pr_err("Access read fail\n");
		return -1;
	}
    if (unlikely(!access_ok(VERIFY_READ, (void __user *)&pServiceUrb->descriptorType, sizeof (pServiceUrb->descriptorType)))){
		pr_err("Access read fail\n");
		return -1;
	}
    __get_user(*size, &pServiceUrb->size);
    __get_user(*hcd_index, &pServiceUrb->descriptorType);
    return 0;
}


int32_t usbip_recv_iso (const void __user *buf, size_t count, struct urb *urb){
	return 0;
}



